from typing import Tuple


class Signature:
    """
    Signature represents time signature which can be used as time signature or as duration.
    """
    __slots__ = ('_nominator', '_denominator')

    def __init__(self, nominator: int, denominator: int):
        """
        Initializes Signature object. Nominator (first argument) must be non-negative.
        Denominator (second argument) should be a positive power of two.

        >>> Signature(0, 1)
        Signature(0, 1)

        >>> Signature(1, 4)
        Signature(1, 4)

        >>> Signature(3, 8)
        Signature(3, 8)

        :param nominator: signature nominator
        :param denominator: signature denominator
        """
        if not (nominator >= 0 and denominator > 0 and (denominator & (denominator - 1) == 0)):
            raise ValueError(
                'Invalid time signature '
                '(the nominator must be non-negative and the denominator must be a positive power of two)'
            )

        self._nominator = nominator
        self._denominator = denominator

    @property
    def nominator(self) -> int:
        """
        Returns the nominator of the signature.

        :return: nominator of the signature
        """
        return self._nominator

    @property
    def denominator(self) -> int:
        """
        Returns the denominator of the signature.

        :return: denominator of the signature
        """
        return self._denominator

    def to(self, denominator: int) -> 'Signature':
        """
        Returns new signature with required denominator and nominator scaled accordingly.
        New denominator must be greater or equal to the current one.

        >>> Signature(1, 2).to(4)
        Signature(2, 4)

        >>> Signature(0, 1).to(8)
        Signature(0, 8)

        >>> Signature(1, 1).to(8)
        Signature(8, 8)

        :param denominator: required new denominator
        :return: new signature with required denominator and nominator scaled accordingly
        """
        if not (denominator > 0 and (denominator & (denominator - 1) == 0)):
            raise ValueError('Invalid denominator (must be a positive power of two)')

        if denominator == self._denominator:
            return Signature(self._nominator, self._denominator)

        if denominator > self._denominator:
            nominator = denominator // self._denominator * self._nominator
            return Signature(nominator, denominator)

        raise ValueError(f'New denominator must be >= {self._denominator}')

    def normalized(self) -> 'Signature':
        """
        Returns normalized copy of this signature Signature is normalized when
        nominator and denominator does not have common divisors.

        >>> Signature(1, 4).normalized()
        Signature(1, 4)

        >>> Signature(2, 4).normalized()
        Signature(1, 2)

        >>> Signature(16, 4).normalized()
        Signature(4, 1)

        :return: normalized copy of this signature.
        """
        n, d = self._nominator, self._denominator

        while n & 1 == 0 and d > 1:
            n >>= 1
            d >>= 1

        return Signature(n, d)

    def _to_common_denominator(self, other: 'Signature') -> Tuple[int, int, int]:
        n1, d1 = self._nominator, self._denominator
        n2, d2 = other._nominator, other._denominator

        if d1 == d2:
            return n1, n2, d1

        if d1 > d2:
            return n1, d1 // d2 * n2, d1

        return n2, d2 // d1 * n1, d2

    def __str__(self) -> str:
        """
        Returns human-readable string representation of the signature.

        :return: human-readable string representation of the signature
        """
        return f'{self._nominator}/{self._denominator}'

    def __repr__(self) -> str:
        """
        Returns string representation of the signature.

        :return: string representation of the signature
        """
        return f'{self.__class__.__name__}({self._nominator}, {self._denominator})'

    def __eq__(self, other: 'Signature') -> bool:
        """
        Compares two signatures for equality.
        Signatures are transformed to the common denominator and then their nominators are compared.

        :param other: other signature
        :return: True if signatures equal, False otherwise
        """
        n1, n2, _ = self._to_common_denominator(other)
        return n1 == n2

    def __le__(self, other: 'Signature') -> bool:
        """
        Compares two signatures.
        Signatures are transformed to the common denominator and then their nominators are compared.

        :param other: other signature
        :return: True if this signature is less or equal than other signature
        """
        n1, n2, _ = self._to_common_denominator(other)
        return n1 <= n2

    def __lt__(self, other: 'Signature') -> bool:
        """
        Compares two signatures.
        Signatures are transformed to the common denominator and then their nominators are compared.

        :param other: other signature
        :return: True if this signature is less than other signature
        """
        n1, n2, _ = self._to_common_denominator(other)
        return n1 < n2

    def __ge__(self, other: 'Signature') -> bool:
        """
        Compares two signatures.
        Signatures are transformed to the common denominator and then their nominators are compared.

        :param other: other signature
        :return: True if this signature is greater or equal than other signature
        """
        n1, n2, _ = self._to_common_denominator(other)
        return n1 >= n2

    def __gt__(self, other: 'Signature') -> bool:
        """
        Compares two signatures.
        Signatures are transformed to the common denominator and then their nominators are compared.

        :param other: other signature
        :return: True if this signature is greater than other signature
        """
        n1, n2, _ = self._to_common_denominator(other)
        return n1 > n2

    def __add__(self, other: 'Signature') -> 'Signature':
        """
        Returns sum of two signatures.
        Before summation signatures are transformed to the common denominator.

        :param other: other signature
        :return: sum of two signatures
        """
        n1, n2, d = self._to_common_denominator(other)
        return Signature(n1 + n2, d)

    def __mul__(self, other: 'Signature') -> 'Signature':
        """
        Returns multiplication of two signatures.

        :param other: other signature
        :return: multiplication of two signatures
        """
        return Signature(self._nominator * other._nominator, self._denominator * other._denominator)

    def __abs__(self) -> 'Signature':
        """
        Returns absolute value of the signature.

        :return: absolute value of the signature
        """
        return Signature(abs(self._nominator), self._denominator)



import re
from typing import Dict

_notation_regex = re.compile(r'(?P<tone>[ABCDEFG])(?P<transposition>[#b]*)(?P<octave>-?\d+)?')
_tone_mapping: Dict[str, int] = {
    'C': 0,
    'D': 2,
    'E': 4,
    'F': 5,
    'G': 7,
    'A': 9,
    'B': 11
}
_reverse_tone_mapping: Dict[int, str] = {v: k for k, v in _tone_mapping.items()}


class Tone:
    """
    Tone represents pitch from chromatic scale.
    """
    __slots__ = ('_pitch',)

    def __init__(self, pitch: int):
        """
        Initializes Tone object. Takes pitch as only argument.
        Pitch must be an integer number corresponding to tone number in chromatic scale.
        For example, C0 is 0, C#0 is 1, D0 is 2, C1 is 12, A4 is 57.

        >>> Tone(0).to_notation()
        'C0'

        >>> Tone(1).to_notation()
        'C#0'

        >>> Tone(57).to_notation()
        'A4'

        If you want to initialize tone from notation (for example from string 'Db3'), use
        `Tone.from_notation` instead.

        :param pitch: any integer number corresponding to tone number in chromatic scale
        """
        self._pitch = pitch

    @property
    def pitch(self) -> int:
        """
        Returns pitch of the tone.

        :return: pitch of the tone
        """
        return self._pitch

    @property
    def octave(self) -> int:
        """
        Returns octave of the tone.

        >>> Tone.from_notation('C5').octave
        5

        >>> Tone.from_notation('F-2').octave
        -2

        :return: octave of the tone
        """

        return self._pitch // 12

    @classmethod
    def from_notation(cls, notation: str) -> 'Tone':
        """
        Parses tone from notation. Multiple transposition characters ('#' or 'b') are allowed.
        Tone name should be uppercase to avoid ambiguity. Octave number can be negative.
        Raises ValueError if notation can not be parsed.

        >>> Tone.from_notation('C')
        Tone(pitch=0)

        >>> Tone.from_notation('A4')
        Tone(pitch=57)

        >>> Tone.from_notation('Db3')
        Tone(pitch=37)

        >>> Tone.from_notation('G#5')
        Tone(pitch=68)

        >>> Tone.from_notation('G#b#bb#5')
        Tone(pitch=67)

        :param notation: notation string
        :return: Tone with pitch derived from notation
        """
        match = _notation_regex.match(notation)

        if match is None:
            raise ValueError('Invalid notation')

        base_pitch = _tone_mapping[match.group('tone')]
        transposition = sum(+1 if t == '#' else -1 for t in match.group('transposition'))
        octave_raw = match.group('octave')
        octave = int(octave_raw) if octave_raw else 0

        return cls(12 * octave + base_pitch + transposition)

    def to_notation(self, transpose_down: bool = False) -> str:
        """
        Formats tone to notation.

        >>> Tone(0).to_notation()
        'C0'

        >>> Tone(13).to_notation()
        'C#1'

        >>> Tone(13).to_notation(transpose_down=True)
        'Db1'

        >>> Tone(113).to_notation()
        'F9'

        >>> Tone(113).to_notation(transpose_down=True)
        'F9'

        :param transpose_down: transpose tones down with flat symbol ('b')
        instead of transposing up with sharp symbol ('#')
        :return: tone notation
        """
        base_pitch = self._pitch % 12
        octave = self._pitch // 12
        if base_pitch in _reverse_tone_mapping:
            tone = _reverse_tone_mapping[base_pitch]
            transposition = ''
        else:
            if transpose_down:
                tone = _reverse_tone_mapping[base_pitch + 1]
                transposition = 'b'
            else:
                tone = _reverse_tone_mapping[base_pitch - 1]
                transposition = '#'

        return f'{tone:s}{transposition:s}{octave:d}'

    def to_frequency(self, base: float = 440.0) -> float:
        """
        Returns frequency of the tone based on the frequency of the A4 tone.

        :param base: frequency of the base tone (A4) in Hz (default: 440.0)
        :return: frequency of the tone in Hz
        """

        return base * 2 ** ((self._pitch - 57) / 12)

    def transposed(self, transposition: int) -> 'Tone':
        """
        Returns transposed copy of the tone.

        >>> Tone.from_notation('C4').transposed(1).to_notation()
        'C#4'

        >>> Tone.from_notation('C4').transposed(2).to_notation()
        'D4'

        >>> Tone.from_notation('C4').transposed(-1).to_notation()
        'B3'

        >>> Tone.from_notation('C4').transposed(-2).to_notation()
        'A#3'

        :param transposition: integer number of semitones to transpose tone
        :return: transposed copy of the tone
        """

        return Tone(pitch=self._pitch + transposition)

    def __str__(self) -> str:
        """
        Returns human-readable string representation of the tone.

        :return: string interpretation of the tone
        """
        return self.to_notation()

    def __repr__(self) -> str:
        """
        Returns string representation of the tone.

        :return: string representation of the tone
        """
        return f'{self.__class__.__name__}(pitch={self._pitch})'

    def __eq__(self, other: 'Tone') -> bool:
        """
        Compares two tones for equality. Tones are equal if their pitches are equal.

        :param other: other tone
        :return: True of tones have equal pitches, False otherwise
        """
        return self._pitch == other._pitch

    def __le__(self, other: 'Tone') -> bool:
        """
        Compares two tones by their pitches.

        :param other: other tone
        :return: True if this pitch tone is less or equal than other tone pitch, False otherwise
        """
        return self._pitch <= other._pitch

    def __lt__(self, other: 'Tone') -> bool:
        """
        Compares two tones by their pitches.

        :param other: other tone
        :return: True if this pitch tone is less than other tone pitch, False otherwise
        """
        return self._pitch < other._pitch

    def __ge__(self, other: 'Tone') -> bool:
        """
        Compares two tones by their pitches.

        :param other: other tone
        :return: True if this pitch tone is greater or equal than other tone pitch, False otherwise
        """
        return self._pitch >= other._pitch

    def __gt__(self, other: 'Tone') -> bool:
        """
        Compares two tones by their pitches.

        :param other: other tone
        :return: True if this pitch tone is greater than other tone pitch, False otherwise
        """
        return self._pitch > other._pitch



import io
from typing import BinaryIO, List, Tuple, Union

from melodia.core.note import Note
from melodia.core.sgnature import Signature
from melodia.core.tone import Tone
from melodia.core.track import Track


class MIDIBase:
    __slots__ = (
        '_pulses_per_quarter',
        '_pulses_per_whole',
        '_middle_c_delta'
    )

    def __init__(
            self,
            pulses_per_quarter: int = 96,
            middle_c: Union[Tone, int, str] = 'C3'
    ):
        self._pulses_per_quarter: int = pulses_per_quarter
        self._pulses_per_whole: int = pulses_per_quarter * 4

        self._middle_c_delta: int
        if isinstance(middle_c, Tone):
            self._middle_c_delta = 60 - middle_c.pitch
        elif isinstance(middle_c, int):
            self._middle_c_delta = 60 - Tone(middle_c).pitch
        elif isinstance(middle_c, str):
            self._middle_c_delta = 60 - Tone.from_notation(middle_c).pitch
        else:
            raise TypeError('middle_c must be a Tone object, an integer or a string')

    def _signature_to_pulses(self, signature: Signature) -> int:
        normalized = signature.normalized()

        return self._pulses_per_whole // normalized.denominator * normalized.nominator

    def _format_note_on(self, note: Note, channel: int) -> bytes:
        assert 0 <= channel <= 15

        status = 0x90 | channel
        pitch = note.tone.pitch + self._middle_c_delta

        if not 0 <= pitch <= 127:
            raise ValueError(f'Tone {str(note.tone)} can not be written to midi file')

        velocity = int(note.velocity * 127)

        assert 0 <= velocity <= 127

        return b''.join([
            status.to_bytes(1, 'big'),
            pitch.to_bytes(1, 'big'),
            velocity.to_bytes(1, 'big')
        ])

    def _format_note_off(self, note: Note, channel: int) -> bytes:
        assert 0 <= channel <= 15

        status = 0x80 | channel
        pitch = note.tone.pitch + self._middle_c_delta

        if not 0 <= pitch <= 127:
            raise ValueError(f'Tone {str(note.tone)} can not be written to midi file')

        velocity = 64

        return b''.join([
            status.to_bytes(1, 'big'),
            pitch.to_bytes(1, 'big'),
            velocity.to_bytes(1, 'big')
        ])

    @staticmethod
    def _format_var_len(x: int) -> bytes:
        buffer = x & 0x7f
        x >>= 7
        while x > 0:
            buffer <<= 8
            buffer |= 0x80
            buffer += (x & 0x7f)
            x >>= 7

        return buffer.to_bytes(buffer.bit_length() // 8 + 1, 'little')

    @staticmethod
    def _format_signature(signature: Signature, cc: int, bb: int) -> bytes:
        nn = signature.nominator
        dd = signature.denominator.bit_length() - 1

        return b''.join([
            b'\xFF\x58\x04',
            nn.to_bytes(1, 'big'),
            dd.to_bytes(1, 'big'),
            cc.to_bytes(1, 'big'),
            bb.to_bytes(1, 'big')
        ])

    @staticmethod
    def _format_tempo(tempo: int) -> bytes:
        return b''.join([
            b'\xFF\x51\x03',
            tempo.to_bytes(3, 'big')
        ])

    @staticmethod
    def _format_chunk(kind: bytes, body: bytes) -> bytes:
        assert len(kind) == 4
        return b''.join([kind, len(body).to_bytes(4, 'big'), body])

    @staticmethod
    def _format_header_chunk(
            kind: int,
            n_tracks: int,
            ticks_per_quarter_note: int
    ):
        assert kind in (0, 1, 2)
        assert n_tracks > 0
        assert ticks_per_quarter_note < 0x7FFF

        body = b''.join([
            kind.to_bytes(2, 'big'),
            n_tracks.to_bytes(2, 'big'),
            ticks_per_quarter_note.to_bytes(2, 'big')
        ])

        return MIDIBase._format_chunk(b'MThd', body)


class MIDIWriter(MIDIBase):
    __slots__ = ('_bpm', '_channel')

    def __init__(
            self,
            pulses_per_quarter: int = 96,
            middle_c: Union[Tone, int, str] = 'C3',
            bpm: float = 120.0,
            channel: int = 0
    ):
        super().__init__(
            pulses_per_quarter=pulses_per_quarter,
            middle_c=middle_c
        )
        self._bpm = bpm
        self._channel = channel

    def dump(self, track: Track, file: BinaryIO) -> None:
        absolute_note_events: List[Tuple[int, bytes]] = []
        for position, note in track:
            note_on_absolute_time = self._signature_to_pulses(position)
            note_on_event = self._format_note_on(note, channel=self._channel)
            note_off_absolute_time = note_on_absolute_time + self._signature_to_pulses(note.duration)
            note_off_event = self._format_note_off(note, channel=self._channel)

            absolute_note_events.append((note_on_absolute_time, note_on_event))
            absolute_note_events.append((note_off_absolute_time, note_off_event))

        absolute_note_events.sort(key=lambda x: x[0])

        delta_events: List[Tuple[int, bytes]] = [
            # Time signature
            (0, self._format_signature(track.signature, cc=self._pulses_per_quarter, bb=8)),
            # Tempo
            (0, self._format_tempo(int(60000000 / self._bpm)))
        ]

        last_absolute_time = 0
        for absolute_time, note_event in absolute_note_events:
            delta_events.append((absolute_time - last_absolute_time, note_event))
            last_absolute_time = absolute_time

        # End of track
        delta_events.append((0, b'\xFF\x2F\x00'))

        # Header
        file.write(self._format_header_chunk(
            kind=0,
            n_tracks=1,
            ticks_per_quarter_note=self._pulses_per_quarter
        ))

        # Track
        track_body: List[bytes] = []
        for delta_time, event in delta_events:
            track_body.append(self._format_var_len(delta_time))
            track_body.append(event)
        file.write(self._format_chunk(b'MTrk', b''.join(track_body)))


def dump(track: Track, file: BinaryIO, bpm: float = 120.0, channel: int = 0) -> None:
    MIDIWriter(bpm=bpm, channel=channel).dump(track, file)


def dumps(track: Track, bpm: float = 120.0, channel: int = 0) -> bytes:
    with io.BytesIO() as file:
        dump(track, file, bpm=bpm, channel=channel)
        return file.getbuffer().tobytes()



import heapq
from typing import Iterable, Tuple, Union, List, Optional

from melodia.core.note import Note
from melodia.core.sgnature import Signature


class Track:
    """
    Track represents collection of notes where each note has its position.
    Two or more notes with different tones can share same position.
    """
    __slots__ = (
        '_signature',
        '_max_denominator',
        '_max_position',
        '_heap',
        '_sorted'
    )

    def __init__(
            self,
            signature: Union[Signature, Tuple[int, int]] = Signature(4, 4),
            content: Optional[Iterable[Tuple[Note, Union[Signature, Tuple[int, int], None]]]] = None,
    ):
        """
        Initializes Track object.

        :param signature: time signature of the track (default: (4, 4))
        :param content: optional content of the track as a list of pairs (note, position) (default: None)
        """
        self._signature: Signature
        if isinstance(signature, Signature):
            self._signature = signature
        elif isinstance(signature, Iterable):
            self._signature = Signature(*signature)
        else:
            raise TypeError('signature must be a Signature object or a pair of integers')

        self._max_denominator = 1
        self._heap: List[Tuple[Signature, Note]] = []
        self._max_position: Signature = Signature(0, 1)

        self._sorted = True

        if content is not None:
            for note, position in content:
                self.add(note, where=position)

    @property
    def signature(self) -> Signature:
        """
        Returns time signature of the track.

        :return: time signature of the track
        """
        return self._signature

    @property
    def length(self) -> Signature:
        """
        Returns length of the track.
        Length of the track is defined as the position after which
        no note will be played. In other words it is the right-most ending of the note.

        :return: length of the track
        """
        return self._max_position

    def add(
            self,
            what: Union[Note, Iterable[Note]],
            where: Union[Signature, Tuple[int, int], None] = None
    ) -> None:
        """
        Adds note or multiple notes to the track to the specified position.
        In the case of the multiple notes, all the notes will be placed in the same position.
        If position is not specified, note or notes will be placed at the end of the track
        (after the right-most ending of the note).

        :param what: one note or iterable of notes
        :param where: position where note or multiple notes will be placed (default: None)
        :return: None
        """
        position: Signature
        if where is None:
            position = self._max_position
        elif isinstance(where, Signature):
            position = where
        elif isinstance(where, Iterable):
            position = Signature(*where)
        else:
            raise TypeError('where must be a Signature object or a pair of integers')

        if isinstance(what, Iterable):
            for x in what:
                self.add(what=x, where=position)
            return

        note: Note = what

        if position.denominator > self._max_denominator:
            self._max_denominator = position.denominator
            self._heap = [(s.to(self._max_denominator), n) for s, n in self._heap]

        position_transformed = position.to(self._max_denominator)

        heapq.heappush(self._heap, (position_transformed, note))
        self._sorted = False

        end_position = position_transformed + note.duration
        if end_position > self._max_position:
            self._max_position = end_position

    def __iter__(self) -> Iterable[Tuple[Signature, Note]]:
        """
        Returns track iterator. Iterator yields pairs (position, note) in order of ascending positions.
        Position is the Signature object and note is the Note object.

        :return: iterator of (position, note) pairs
        """
        if not self._sorted:
            self._heap.sort()
            self._sorted = True

        return iter(self._heap)

    def __str__(self) -> str:
        """
        Returns human-readable string representation of the track.

        :return: human-readable string representation of the track
        """
        return f'Track {str(self._signature)} with {len(self._heap)} notes'

    def __repr__(self) -> str:
        """
        Returns string representation of the track.

        :return: string representation of the track
        """
        name = self.__class__.__name__
        signature = f'({self._signature.nominator}, {self._signature.denominator})'

        contents = []
        for position, note in self:
            contents.append(f'({repr(note)}, ({position.nominator}, {position.denominator}))')

        return f'{name}(signature={signature}, content=[{", ".join(contents)}])'



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='melodia',
    version='1.0',
    description='Python library for MIDI music creation',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Egor Malykh',
    author_email='hello@meownoid.pro',
    url='https://github.com/meownoid/melodia',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: MIDI',
        'Typing :: Typed'
    ],
    keywords=[
        'music',
        'midi',
        'note',
        'track',
        'composer',
        'daw',
        'audio'
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'}
)



from typing import Tuple, Union, Iterable

from melodia.core.sgnature import Signature
from melodia.core.tone import Tone


class Note:
    """
    Note represents musical note with three main characteristics: tone, velocity and duration.
    """
    __slots__ = ('_tone', '_velocity', '_duration')

    def __init__(
            self,
            tone: Union[Tone, int, str],
            duration: Union[Signature, Tuple[int, int]] = Signature(1, 4),
            velocity: float = 0.75
    ):
        """
        Initializes Note object. Tone can be a Tone object, an integer or a string.
        In the case of an integer, Tone constructor is called.
        In the case of a string, Tone.from_notation is called.
        Duration can be a Signature object or a pair of integers.
        In the case of a pair of integers Signature constructor is called.
        Velocity must be a float in the range [0.0, 1.0] where 0.0 is the minimum velocity and
        1.0 is the maximum velocity. By default velocity is 0.75 (as in many DAWs).

        :param tone: tone of the note, must be Tone object, integer or string
        :param duration: duration of the note, must be Signature or pair of integers (default: (1, 4))
        :param velocity: float number in the range [0.0, 1.0] (default: 0.75)
        """
        self._tone: Tone
        if isinstance(tone, Tone):
            self._tone = tone
        elif isinstance(tone, int):
            self._tone = Tone(tone)
        elif isinstance(tone, str):
            self._tone = Tone.from_notation(tone)
        else:
            raise TypeError('tone must be a Tone object, an integer or a string')

        self._duration: Signature
        if isinstance(duration, Signature):
            self._duration = duration
        elif isinstance(duration, Iterable):
            self._duration = Signature(*duration)
        else:
            raise TypeError('duration must be a Signature object or a pair of integers')

        if not 0.0 <= velocity <= 1.0:
            raise ValueError('velocity must be in range [0.0, 1.0]')

        self._velocity: float = velocity

    @property
    def tone(self) -> Tone:
        """
        Returns tone of the note.

        :return: tone of the note
        """
        return self._tone

    @property
    def velocity(self) -> float:
        """
        Returns velocity of the note.

        :return: velocity of the note.
        """
        return self._velocity

    @property
    def duration(self) -> Signature:
        """
        Returns duration of the note.

        :return: duration of the note
        """
        return self._duration

    def transposed(self, transposition: int) -> 'Note':
        """
        Returns note with transposed tone. Velocity and duration remains the same.

        :param transposition: integer number of semitones to transpose tone
        :return: note with transposed tone
        """
        return Note(
            tone=self._tone.transposed(transposition),
            velocity=self._velocity,
            duration=self._duration
        )

    def _as_triplet(self) -> Tuple[Tone, Signature, float]:
        return self._tone, self._duration, self._velocity

    def __str__(self) -> str:
        """
        Returns human-readable representation of the note.

        :return: human-readable representation of the note
        """
        return f'Note {str(self._duration)} {str(self._tone)} ({self._velocity:.3f})'

    def __repr__(self) -> str:
        """
        Returns string representation of the note.

        :return: string representation of the note
        """
        name = self.__class__.__name__
        pitch = self._tone.pitch
        velocity = self._velocity
        duration = f'({self._duration.nominator}, {self._duration.denominator})'

        return f'{name}(tone={pitch}, duration={duration}, velocity={velocity}) '

    def __eq__(self, other: 'Note') -> bool:
        """
        Compares two notes for equality. Notes are equal if all their components are equal.

        :param other: other note
        :return: True if notes are equal, False otherwise
        """
        return self._as_triplet() == other._as_triplet()

    def __le__(self, other: 'Note') -> bool:
        """
        Compares two notes. Notes are compared as triplets (pitch, duration, velocity).

        :param other: other note
        :return: True if this note is less or equal than other note, False otherwise
        """
        return self._as_triplet() <= other._as_triplet()

    def __lt__(self, other: 'Note') -> bool:
        """
        Compares two notes. Notes are compared as triplets (pitch, duration, velocity).

        :param other: other note
        :return: True if this note is less than other note, False otherwise
        """
        return self._as_triplet() < other._as_triplet()

    def __ge__(self, other: 'Note') -> bool:
        """
        Compares two notes. Notes are compared as triplets (pitch, duration, velocity).

        :param other: other note
        :return: True if this note is greater or equal than other note, False otherwise
        """
        return self._as_triplet() >= other._as_triplet()

    def __gt__(self, other: 'Note') -> bool:
        """
        Compares two notes. Notes are compared as triplets (pitch, duration, velocity).

        :param other: other note
        :return: True if this note is greater than other note, False otherwise
        """
        return self._as_triplet() > other._as_triplet()
